"""Sync metadata from external sources."""

import logging

import moe
from moe import config
from moe.library import LibItem

log = logging.getLogger("moe.sync")

__all__ = ["sync_item"]


class Hooks:
    """Sync plugin hook specifications."""

    @staticmethod
    @moe.hookspec
    def sync_metadata(item: LibItem):
        """Implement specific metadata syncing for ``item``."""


def sync_item(item: LibItem):
    """Syncs metadata from external sources and merges changes into ``item``."""
    log.debug(f"Syncing metadata. [{item=!r}]")

    config.CONFIG.pm.hook.sync_metadata(item=item)

    log.debug(f"Synced metadata. [{item=!r}]")



"""Core interface for editing an item."""

import datetime
import logging

import sqlalchemy.ext.associationproxy

from moe.library import LibItem

__all__ = ["EditError", "edit_item"]

log = logging.getLogger("moe.edit")


class EditError(Exception):
    """Error editing an item in the library."""


def edit_item(item: LibItem, field: str, value: str):
    """Sets a LibItem's ``field`` to ``value``.

    Args:
        item: Library item to edit.
        field: Item field to edit.
        value: Value to set the item's field to.

    Raises:
        EditError: ``field`` is not a valid attribute or is not editable.
    """
    log.debug(f"Editing item. [{item=!r}, {field=!r}, {value=!r}]")

    try:
        attr = getattr(item, field)
    except AttributeError as a_err:
        raise EditError(f"Invalid field given. [{field=!r}]") from a_err

    non_editable_fields = ["path"]
    if field in non_editable_fields:
        raise EditError(f"Non-editable field given. [{field=!r}]")

    if isinstance(attr, str):
        setattr(item, field, value)
    elif isinstance(attr, int):
        setattr(item, field, int(value))
    elif isinstance(attr, set) or isinstance(
        attr, sqlalchemy.ext.associationproxy._AssociationSet
    ):
        setattr(item, field, {value.strip() for value in value.split(";")})
    elif isinstance(attr, datetime.date):
        try:
            setattr(item, field, datetime.date.fromisoformat(value))
        except ValueError as v_err:
            raise EditError("Date must be in format YYYY-MM-DD") from v_err
    else:
        raise EditError(f"Editing field not supported. [{field=!r}]")

    log.info(f"Item edited. [{item=!r}, {field=!r}, {value=!r}]")



"""Integrates musicbrainz with the import prompt.

The ``musicbrainz`` cli plugin provides the following functionality:

* Ability to search for a specific musicbrainz ID when importing an item.
* `mbcol` command to sync a musicbrainz collection with items in the library.
"""


import argparse
import logging
from typing import Optional

import questionary

import moe
import moe.cli
from moe.library import Album, Extra, Track
from moe.plugins import moe_import
from moe.plugins import musicbrainz as moe_mb
from moe.util.cli import PromptChoice, cli_query, query_parser

__all__: list[str] = []

log = logging.getLogger("moe.cli.mb")


@moe.hookimpl
def add_command(cmd_parsers: argparse._SubParsersAction):
    """Adds the ``mbcol`` command to Moe's CLI."""
    mbcol_parser = cmd_parsers.add_parser(
        "mbcol",
        description="Set a musicbrainz collection to a query.",
        help="sync a musicbrainz collection",
        parents=[query_parser],
    )
    col_option_group = mbcol_parser.add_mutually_exclusive_group()
    col_option_group.add_argument(
        "--add",
        action="store_true",
        help="add items to a collection",
    )
    col_option_group.add_argument(
        "--remove",
        action="store_true",
        help="remove items from a collection",
    )
    mbcol_parser.set_defaults(func=_parse_args)


def _parse_args(args: argparse.Namespace):
    """Parses the given commandline arguments.

    Args:
        args: Commandline arguments to parse.

    Raises:
        SystemExit: Invalid query given, or no items to remove.
    """
    items = cli_query(args.query, query_type=args.query_type)

    releases = set()
    for item in items:
        release_id: Optional[str] = None
        if isinstance(item, (Extra, Track)):
            release_id = item.album_obj.mb_album_id
        elif isinstance(item, Album):
            release_id = item.mb_album_id

        if release_id:
            releases.add(release_id)

    if not releases:
        log.error("Queried items don't contain any musicbrainz releases to sync.")
        raise SystemExit(1)

    if args.add:
        moe_mb.add_releases_to_collection(releases)
    elif args.remove:
        moe_mb.rm_releases_from_collection(releases)
    else:
        moe_mb.set_collection(releases)


@moe.hookimpl
def add_import_prompt_choice(prompt_choices: list[PromptChoice]):
    """Adds a choice to the import prompt to allow specifying a mb id."""
    prompt_choices.append(
        PromptChoice(title="Enter Musicbrainz ID", shortcut_key="m", func=_enter_id)
    )


def _enter_id(
    old_album: Album,
    new_album: Album,
):
    """Re-run the add prompt with the inputted Musibrainz release."""
    mb_id = questionary.text("Enter Musicbrainz ID: ").ask()

    log.debug(
        f"Re-running import prompt for different musicbrainz release. [{mb_id=!r}]"
    )

    album = moe_mb.get_album_by_id(mb_id)

    moe_import.import_prompt(old_album, album)



"""This package contains shared functionality for the CLI."""

from . import prompt, query
from .prompt import *
from .query import *

__all__ = []
__all__.extend(prompt.__all__)
__all__.extend(query.__all__)



"""Moe database/library functionality."""

from . import album, extra, lib_item, track
from .album import *
from .extra import *
from .lib_item import *
from .track import *

__all__ = []
__all__.extend(album.__all__)
__all__.extend(extra.__all__)
__all__.extend(lib_item.__all__)
__all__.extend(track.__all__)



"""Detect and handle duplicates in the library."""

import logging
from typing import Optional

import sqlalchemy

import moe
from moe import config
from moe.library import Album, Extra, LibItem, Track
from moe.query import query

__all__ = ["get_duplicates"]

log = logging.getLogger("moe.dup")


class DuplicateError(Exception):
    """Duplicate items could not be resolved."""


class Hooks:
    """Duplicate plugin hook specifications."""

    @staticmethod
    @moe.hookspec
    def resolve_dup_items(item_a: LibItem, item_b: LibItem):
        """Resolve two duplicate items.

        A resolution should come in one of two forms:
        1. The field(s) triggering the duplicate detection is altered on one or both
           of the items so that they are no longer duplicates.
        2. One of the items is removed from the library entirely using
           :meth:`~moe.plugins.remove.remove_item`.

        Important:
            Duplicates are determined by each item's
            :meth:`~moe.library.lib_item.LibItem.is_unique` method. This behavior can be
            extended by the various ``is_unique_*`` hooks.

        Note:
            As a plugin that may have introduced new ``unique`` constraints via one of
            the ``is_unique_*`` hooks, you only need to be concerned about resolving
            that particular constraint. You may also just rely on the default duplicate
            resolution method of Moe, which depending on the UI in use, may be
            sufficient for your plugin's needs. For example, the CLI will offer a
            'duplicate resolution' prompt to the user anytime a duplicate is detected.
            This hook allows you to implement your own duplicate resolution methods in
            addition to what's offered by default.

        Args:
            item_a: First item.
            item_b: Second item.

        See Also:
            * The :meth:`~moe.library.album.Hooks.is_unique_album` hook.
            * The :meth:`~moe.library.extra.Hooks.is_unique_extra` hook.
            * The :meth:`~moe.library.track.Hooks.is_unique_track` hook.
        """


@moe.hookimpl(hookwrapper=True)
def edit_changed_items(items: list[LibItem]):
    """Check for and resolve duplicates when items are edited."""
    yield  # run all `edit_changed_items` hook implementations
    albums = [item for item in items if isinstance(item, Album)]  # resolve albums first
    tracks = [item for item in items if isinstance(item, Track)]
    extras = [item for item in items if isinstance(item, Extra)]
    resolve_duplicates(albums)  # type: ignore
    resolve_duplicates(tracks)  # type: ignore
    resolve_duplicates(extras)  # type: ignore


@moe.hookimpl(hookwrapper=True)
def edit_new_items(items: list[LibItem]):
    """Check for and resolve duplicates when items are added to the library."""
    yield  # run all `edit_new_items` hook implementations
    albums = [item for item in items if isinstance(item, Album)]  # resolve albums first
    tracks = [item for item in items if isinstance(item, Track)]
    extras = [item for item in items if isinstance(item, Extra)]
    resolve_duplicates(albums)  # type: ignore
    resolve_duplicates(tracks)  # type: ignore
    resolve_duplicates(extras)  # type: ignore


def resolve_duplicates(items: list[LibItem]):
    """Search for and resolve any duplicates of items in ``items``."""
    log.debug(f"Checking for duplicate items. [{items=!r}]")

    resolved_items = []
    for item in items:
        if _is_removed(item):
            continue

        dup_items = get_duplicates(item, items)
        dup_items += get_duplicates(item)

        for dup_item in dup_items:
            if dup_item in resolved_items or _is_removed(item):
                continue

            log.debug(
                f"Resolving duplicate items. [item_a={item!r}, item_b={dup_item!r}]"
            )
            config.CONFIG.pm.hook.resolve_dup_items(item_a=item, item_b=dup_item)
            if (
                not item.is_unique(dup_item)
                and not _is_removed(item)
                and not _is_removed(dup_item)
            ):
                raise DuplicateError(
                    "Duplicate items could not be resolved. "
                    f"[item_a={item!r}, item_b={dup_item!r}]"
                )

        if dup_items:
            resolved_items.append(item)

    if resolved_items:
        log.debug(f"Resolved duplicate items. [{resolved_items=!r}]")
    else:
        log.debug("No duplicate items found.")


def _is_removed(item):
    """Check whether or not an item is removed from the library."""
    insp = sqlalchemy.inspect(item)

    if insp.deleted or insp.transient:
        return True

    return False


def get_duplicates(
    item: LibItem, others: Optional[list[LibItem]] = None
) -> list[LibItem]:
    """Returns items considered duplicates of ``item``.

    Args:
        item: Library item to get duplicates of.
        others: Items to compare against. If not given, will query the database
            and compare against all items in the library.

    Returns:
        Any items considered a duplicate as defined by
        :meth:`~moe.library.lib_item.LibItem.is_unique`.
    """
    dup_items = []
    if not others:
        others = query("*", type(item).__name__.lower())

    for other in others:
        if (
            item is not other
            and type(item) == type(other)
            and not item.is_unique(other)
        ):
            dup_items.append(other)

    return dup_items



#!/usr/bin/env python3

"""Entry point for the CLI.

This module deals with parsing the arguments of ``moe`` and related functionality.
For general shared CLI functionality, see the ``moe.util.cli`` package.
"""

import argparse
import logging
import sys

import pkg_resources
import pluggy
from rich.console import Console

import moe
from moe import config
from moe.config import Config, ConfigValidationError, MoeSession

__all__ = ["console"]

log = logging.getLogger("moe.cli")

console = Console()


class Hooks:
    """General CLI hook specifications."""

    @staticmethod
    @moe.hookspec
    def add_command(cmd_parsers: argparse._SubParsersAction):
        """Add a sub-command to Moe's CLI.

        Args:
            cmd_parsers: Contains all the sub-command parsers. The new sub-command
                should be added as an argparse parser to `cmd_parsers`.

        Example:
            Inside your hook implementation::

                my_parser = cmd_parsers.add_parser('<command_name>', help='')
                my_parser.add_argument('bar', type=int)
                my_parser.set_defaults(func=my_function)

        Important:
            To specify a function to run when your command is passed, you need to
            define the ``func`` key using ``set_defaults`` as shown above.
            The function will be called like::

                func(
                    config: moe.Config,  # user config
                    args: argparse.Namespace,  # parsed commandline arguments
                )

        Note:
            If your command utilizes a `query`, you can specify ``query_parser`` as
            a parent parser::

                edit_parser = cmd_parsers.add_parser(
                    "list",
                    description="Lists music in the library.",
                    parents=[moe.cli.query_parser],
                )

            Then, you can call ``query.query(args.query, query_type=args.query_type)``
            to get a list of items matching the query from the library.

        See Also:
            * `The python documentation for adding sub-parsers.
              <https://docs.python.org/3/library/argparse.html#sub-commands>`_
            * The :meth:`~moe.query.query` function.
        """


@moe.hookimpl
def add_hooks(pm: pluggy.manager.PluginManager):
    """Registers `CLI` hookspecs to Moe."""
    from moe.cli import Hooks

    pm.add_hookspecs(Hooks)


def main(args: list[str] = sys.argv[1:]):
    """Runs the CLI."""
    log.debug(f"Commandline arguments received. [{args=!r}]")

    if not config.CONFIG:
        try:
            Config()
        except ConfigValidationError as err:
            log.error(err)
            raise SystemExit(1) from err
    _parse_args(args)


if __name__ == "__main__":
    main()


def _parse_args(args: list[str]):
    """Parses the commandline arguments.

    Args:
        args: Arguments to parse. Should not include 'moe'.

    Raises:
        SystemExit: No sub-commands given.
            Does not include root commands such as `--version` or `--help`.
    """
    moe_parser = _create_arg_parser()

    # load all sub-commands
    cmd_parsers = moe_parser.add_subparsers(help="command to run", dest="command")
    config.CONFIG.pm.hook.add_command(cmd_parsers=cmd_parsers)

    parsed_args = moe_parser.parse_args(args)

    # no sub-command given
    if not parsed_args.command:
        moe_parser.print_help(sys.stderr)
        raise SystemExit(1)

    _set_log_lvl(parsed_args)

    # call the sub-command's handler within a single session
    cli_session = MoeSession()
    with cli_session.begin():
        try:
            parsed_args.func(args=parsed_args)
        except SystemExit:
            cli_session.commit()
            raise


def _create_arg_parser() -> argparse.ArgumentParser:
    """Creates the root argument parser."""
    version = pkg_resources.get_distribution("moe").version

    moe_parser = argparse.ArgumentParser()
    moe_parser.add_argument(
        "--version", action="version", version=f"%(prog)s v{version}"
    )
    moe_parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        help="increase logging verbosity; use -vv to enable debug logging",
    )
    moe_parser.add_argument(
        "--quiet",
        "-q",
        action="count",
        help="decrease logging verbosity; use -qq to limit logging to critical errors",
    )

    return moe_parser


def _set_log_lvl(args):
    """Sets the root logger level based on cli arguments."""
    moe_log = logging.getLogger("moe")

    if args.verbose == 1:
        moe_log.setLevel(logging.INFO)
    elif args.verbose == 2:
        moe_log.setLevel(logging.DEBUG)
    elif args.quiet == 1:
        moe_log.setLevel(logging.ERROR)
    elif args.quiet == 2:
        moe_log.setLevel(logging.CRITICAL)



"""Import prompt."""

import logging
from typing import Optional, Union

import pluggy
from rich import box
from rich.console import Group
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

import moe
import moe.cli
from moe import config
from moe.cli import console
from moe.library import Album, Track
from moe.util.cli import PromptChoice, choice_prompt
from moe.util.core import get_matching_tracks

log = logging.getLogger("moe.cli.import")

__all__ = ["AbortImport", "import_prompt"]


class AbortImport(Exception):
    """Used to abort the import process."""


class Hooks:
    """Import plugin cli hook specifications."""

    @staticmethod
    @moe.hookspec
    def add_import_prompt_choice(prompt_choices: list[PromptChoice]):
        """Add a user input choice to the import prompt.

        ``func`` should return the album to be added to the library (or ``None`` if no
        album should be added) and will be supplied the following keyword arguments:

            * ``old_album``: Old album with no changes applied.
            * ``new_album``: New album consisting of all the new changes.

        Args:
            prompt_choices: List of prompt choices. To add a prompt choice, simply
                append it to this list.

        Example:
            .. code:: python

                prompt_choices.append(
                    PromptChoice(
                        title="Abort", shortcut_key="x", func=_abort_changes
                    )
                )
        """


@moe.hookimpl
def add_hooks(pm: pluggy.manager.PluginManager):
    """Registers `import` cli hookspecs to Moe."""
    from moe.plugins.moe_import.import_cli import Hooks

    pm.add_hookspecs(Hooks)


@moe.hookimpl
def process_candidates(old_album: Album, candidates):
    """Use the import prompt to select and process the imported candidate albums."""
    if candidates:
        chosen_candidate = candidates[0]
        log.debug(
            "Candidate album chosen for import prompt. "
            f"[candidate={chosen_candidate!r}]"
        )
        try:
            import_prompt(old_album, chosen_candidate)
        except AbortImport as err:
            log.debug(err)
            raise SystemExit(0) from err


@moe.hookimpl
def add_import_prompt_choice(prompt_choices: list[PromptChoice]):
    """Adds the ``apply`` and ``abort`` prompt choices to the user prompt."""
    prompt_choices.append(
        PromptChoice(title="Apply changes", shortcut_key="a", func=_apply_changes)
    )
    prompt_choices.append(
        PromptChoice(title="Abort", shortcut_key="x", func=_abort_changes)
    )


def import_prompt(
    old_album: Album,
    new_album: Album,
):
    """Runs the interactive prompt for the given album changes.

    Args:
        old_album: Album to be added. Any changes will be applied to this album.
        new_album: New album with all metadata changes. Will be compared against
            ``old_album``.

    Raises:
        AbortImport: Import prompt was aborted by the user.
    """
    log.debug(f"Running import prompt. [{old_album=!r}, {new_album=!r}]")

    console.print(_fmt_import_updates(old_album, new_album))

    prompt_choices: list[PromptChoice] = []
    config.CONFIG.pm.hook.add_import_prompt_choice(prompt_choices=prompt_choices)

    prompt_choice = choice_prompt(prompt_choices)
    prompt_choice.func(old_album, new_album)


def _apply_changes(
    old_album: Album,
    new_album: Album,
):
    """Applies the album changes."""
    log.debug("Applying changes from import prompt.")

    for old_track, new_track in get_matching_tracks(old_album, new_album):
        if not old_track and new_track:
            new_album.tracks.remove(new_track)  # missing track
        elif old_track and not new_track:
            old_album.tracks.remove(old_track)  # unmatched track
        elif (
            old_track
            and new_track
            and old_album.get_track(new_track.track_num, new_track.disc) != old_track
        ):
            # matchup track and disc numbers of matches to ensure they merge properly
            old_track.track_num = new_track.track_num
            old_track.disc = new_track.disc

    old_album.merge(new_album, overwrite=True)


def _abort_changes(
    old_album: Album,
    new_album: Album,
):
    """Aborts the album changes."""
    raise AbortImport("Import prompt aborted; no changes made.")


def _fmt_import_updates(old_album: Album, new_album: Album) -> Panel:
    """Formats import updates from `old_album` to `new_album`."""
    album_text = _fmt_album(old_album, new_album)
    track_text = _fmt_tracks(old_album, new_album)

    return Panel(
        Group(album_text, track_text),
        title="Import Updates",
        border_style="light_cyan1",
    )


def _fmt_album(old_album: Album, new_album: Album) -> Text:
    """Formats the header for the album changes panel."""
    header_text = Text(justify="center", style="bold")

    for header_field in ("title", "artist"):
        field_changes = _fmt_field_changes(old_album, new_album, header_field)
        if not field_changes:
            field_changes = Text(getattr(old_album, header_field))

        header_text.append_text(field_changes).append("\n")

    sub_header_text = Text()
    for sub_header in ("media", "year", "country", "label"):
        field_changes = _fmt_field_changes(old_album, new_album, sub_header)
        if not field_changes:
            if getattr(old_album, sub_header):
                field_changes = Text(getattr(old_album, sub_header))

        if sub_header_text and field_changes:
            sub_header_text.append(" | ")
        if field_changes:
            sub_header_text.append_text(field_changes)

    return header_text.append_text(sub_header_text).append("\n")


def _fmt_tracks(old_album: Album, new_album: Album) -> Table:
    """Formats the tracklist differences between two albums."""
    track_table = Table(box=box.SIMPLE)
    track_table.add_column("status")
    track_table.add_column("disc")
    track_table.add_column("#")
    track_table.add_column("Artist")
    track_table.add_column("Title")

    matches = get_matching_tracks(old_album, new_album)
    matches.sort(
        key=lambda match: (
            getattr(match[1], "disc", 0),
            getattr(match[1], "track_num", 0),
        )
    )  # sort by new track's disc then track number

    unmatched_tracks: list[Track] = []
    missing_tracks: list[Track] = []
    for old_track, new_track in matches:
        if old_track and new_track:
            track_table.add_row(
                Text("matched", style="green"),
                _fmt_field_changes(old_track, new_track, "disc"),
                _fmt_field_changes(old_track, new_track, "track_num"),
                _fmt_field_changes(old_track, new_track, "artist"),
                _fmt_field_changes(old_track, new_track, "title"),
            )
        elif old_track and not new_track:
            unmatched_tracks.append(old_track)
        elif not old_track and new_track:
            missing_tracks.append(new_track)
    track_table.rows[-1].end_section = True

    for missing_track in sorted(missing_tracks):
        track_table.add_row(
            Text("missing", style="red"),
            str(missing_track.disc),
            str(missing_track.track_num),
            missing_track.artist,
            missing_track.title,
        )
    for unmatched_track in sorted(unmatched_tracks):
        track_table.add_row(
            Text("unmatched", style="red"),
            str(unmatched_track.disc),
            str(unmatched_track.track_num),
            unmatched_track.artist,
            unmatched_track.title,
        )

    return track_table


def _fmt_field_changes(
    old_item: Union[Album, Track], new_item: Union[Album, Track], field: str
) -> Optional[Text]:
    """Formats changes of a single field.

    Args:
        old_item: Old track or album to compare.
        new_item: New track or album to compare.
        field: Field to compare.

    Returns:
        A rich 'Text' based on the following cases:

        1. `old_item.field` DNE and `new_item.field` exists:
        `new_item.field` will be returned in green
        2. `old_item.field` exists and `new_item.field` DNE:
        `old_item.field`  will be striketroughed and in red
        3. `old_item.field` differs from `new_item.field`:
            "{old_item.field} -> {new_item.field}" and the arrow is yellow
        4. `old_item.field` is equal to `new_item.field`:
           `old_item.field` returned
        5. None if `old_item.field` and `new_item.field` DNE
    """
    old_field = getattr(old_item, field)
    new_field = getattr(new_item, field)

    if not old_field and new_field:
        return Text(str(new_field), style="green")
    if old_field and not new_field:
        return Text(str(old_field), style="red strike")
    if old_field != new_field:
        return (
            Text(str(old_field))
            .append(Text(" -> ", style="yellow"))
            .append(Text(str(new_field)))
        )

    if old_field:
        return Text(str(old_field))
    else:
        return None



"""Adds music to the library.

This module provides the main entry point into the add process via ``add_item()``.
"""

import logging

import pluggy

import moe
from moe import config
from moe.library import LibItem

__all__ = ["AddAbortError", "AddError", "add_item"]

log = logging.getLogger("moe.add")


class Hooks:
    """Add plugin hook specifications."""

    @staticmethod
    @moe.hookspec
    def pre_add(item: LibItem):
        """Provides an item prior to it being added to the library.

        Use this hook if you wish to change an item's metadata prior to it being
        added to the library.

        Args:
            item: Library item being added.

        Note:
            Any UI application should have a way of detecting and resolving duplicate
            items prior to them being added to the database. You may consider
            implementing a ``hookwrapper`` to run conflict resolution code after the
            ``pre_add`` hook is complete, but before the item has been added to the db.

        See Also:
            * The :meth:`~moe.library.lib_item.Hooks.edit_new_items` hook for editing
              items that have been added
              to the library.
            * `Pluggy hook wrapper documention
              <https://pluggy.readthedocs.io/en/stable/#wrappers>`_
        """


@moe.hookimpl
def add_hooks(pm: pluggy.manager.PluginManager):
    """Registers `add` hookspecs to Moe."""
    from moe.plugins.add.add_core import Hooks

    pm.add_hookspecs(Hooks)


class AddError(Exception):
    """Error adding an item to the library."""


class AddAbortError(Exception):
    """Add process has been aborted by the user."""


def add_item(item: LibItem):
    """Adds a LibItem to the library.

    Args:
        item: Item to be added.

    Raises:
        AddError: Unable to add the item to the library.
    """
    log.debug(f"Adding item to the library. [{item=!r}]")
    session = config.MoeSession()

    config.CONFIG.pm.hook.pre_add(item=item)
    session.add(item)
    session.flush()

    log.info(f"Item added to the library. [{item=!r}]")



"""Any non-music item attached to an album such as log files are considered extras."""

import logging
from pathlib import Path, PurePath
from typing import Any, cast

import pluggy
from sqlalchemy import JSON, Column, Integer
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import relationship
from sqlalchemy.schema import ForeignKey

import moe
from moe import config
from moe.library.album import Album
from moe.library.lib_item import LibItem, PathType, SABase

__all__ = ["Extra"]

log = logging.getLogger("moe.extra")


class Hooks:
    """Extra hook specifications."""

    @staticmethod
    @moe.hookspec
    def create_custom_extra_fields() -> dict[str, Any]:  # type: ignore
        """Creates new custom fields for an Extra.

        Returns:
            Dict of the field names to their default values or ``None`` for no default.

        Example:
            .. code:: python

                return {"my_new_field": "default value", "other_field": None}

            You can then access your new field as if it were a normal field::

                extra.my_new_field = "awesome new value"

        Important:
            Your custom field should follow the same naming rules as any other python
            variable i.e. no spaces, starts with a letter, and consists solely of
            alpha-numeric and underscore characters.
        """  # noqa: DAR202

    @staticmethod
    @moe.hookspec
    def is_unique_extra(extra: "Extra", other: "Extra") -> bool:  # type: ignore
        """Add new conditions to determine whether two extras are unique.

        "Uniqueness" is meant in terms of whether the two extras should be considered
        duplicates in the library. These additional conditions will be applied inside a
        extra's :meth:`is_unique` method.
        """


@moe.hookimpl
def add_hooks(pm: pluggy.manager.PluginManager):
    """Registers `extra` hookspecs to Moe."""
    from moe.library.extra import Hooks

    pm.add_hookspecs(Hooks)


class Extra(LibItem, SABase):
    """An Album can have any number of extra files such as logs, cues, etc.

    Attributes:
        album_obj (Album): Album the extra file belongs to.
        path (pathlib.Path): Filesystem path of the extra file.
    """

    __tablename__ = "extra"

    _id: int = cast(int, Column(Integer, primary_key=True))
    path: Path = cast(Path, Column(PathType, nullable=False, unique=True))
    _custom_fields: dict[str, Any] = cast(
        dict[str, Any],
        Column(
            MutableDict.as_mutable(JSON(none_as_null=True)),
            default="{}",
            nullable=False,
        ),
    )

    _album_id: int = cast(int, Column(Integer, ForeignKey("album._id")))
    album_obj: Album = relationship("Album", back_populates="extras")

    def __init__(self, album: Album, path: Path, **kwargs):
        """Creates an Extra.

        Args:
            album: Album the extra file belongs to.
            path: Filesystem path of the extra file.
            **kwargs: Any other fields to assign to the extra.
        """
        self._custom_fields = self._get_default_custom_fields()
        self._custom_fields_set = set(self._custom_fields)

        album.extras.append(self)
        self.path = path

        for key, value in kwargs.items():
            setattr(self, key, value)

        log.debug(f"Extra created. [extra={self!r}]")

    @property
    def fields(self) -> set[str]:
        """Returns any editable, extra fields."""
        return {"album_obj", "path"}.union(self._custom_fields)

    @property
    def rel_path(self) -> PurePath:
        """Returns the extra's path relative to its album's path."""
        return self.path.relative_to(self.album_obj.path)

    def is_unique(self, other: "Extra") -> bool:
        """Returns whether an extra is unique in the library from ``other``."""
        if self.path == other.path:
            return False

        custom_uniqueness = config.CONFIG.pm.hook.is_unique_extra(
            extra=self, other=other
        )
        if False in custom_uniqueness:
            return False

        return True

    def merge(self, other: "Extra", overwrite: bool = False):
        """Merges another extra into this one.

        Args:
            other: Other extra to be merged with the current extra.
            overwrite: Whether or not to overwrite self if a conflict exists.
        """
        log.debug(
            f"Merging extras. [extra_a={self!r}, extra_b={other!r}, {overwrite=!r}]"
        )

        omit_fields = {"album_obj"}
        for field in self.fields - omit_fields:
            other_value = getattr(other, field)
            self_value = getattr(self, field)
            if other_value and (overwrite or (not overwrite and not self_value)):
                setattr(self, field, other_value)

        log.debug(
            f"Extras merged. [extra_a={self!r}, extra_b={other!r}, {overwrite=!r}]"
        )

    def __eq__(self, other):
        """Compares Extras by their fields."""
        if not isinstance(other, Extra):
            return False

        for field in self.fields:
            if not hasattr(other, field) or (
                getattr(self, field) != getattr(other, field)
            ):
                return False

        return True

    def __lt__(self, other: "Extra") -> bool:
        """Sort based on album then path."""
        if self.album_obj == other.album_obj:
            return self.path < other.path

        return self.album_obj < other.album_obj

    def __repr__(self):
        """Represents an Extra using its path and album."""
        field_reprs = []
        omit_fields = {"album_obj"}
        for field in self.fields - omit_fields:
            if hasattr(self, field):
                field_reprs.append(f"{field}={getattr(self, field)!r}")
        repr_str = "Extra(" + ", ".join(field_reprs) + f", album='{self.album_obj}'"

        custom_field_reprs = []
        for custom_field, value in self._custom_fields.items():
            custom_field_reprs.append(f"{custom_field}={value}")
        if custom_field_reprs:
            repr_str += ", custom_fields=[" + ", ".join(custom_field_reprs) + "]"

        repr_str += ")"
        return repr_str

    def __str__(self):
        """String representation of an Extra."""
        return f"{self.album_obj}: {self.rel_path}"

    def _get_default_custom_fields(self) -> dict[str, Any]:
        """Returns the default custom extra fields."""
        return {
            field: default_val
            for plugin_fields in config.CONFIG.pm.hook.create_custom_extra_fields()
            for field, default_val in plugin_fields.items()
        }



"""new field: label.

Revision ID: 817440447857
Revises: 30668259fcd6
Create Date: 2022-10-06 20:39:23.733361

"""
import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision = "817440447857"
down_revision = "30668259fcd6"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("album", schema=None) as batch_op:
        batch_op.add_column(sa.Column("label", sa.String(), nullable=True))


def downgrade():
    with op.batch_alter_table("album", schema=None) as batch_op:
        batch_op.drop_column("label")



"""Core api for importing albums."""

import logging

import pluggy

import moe
from moe import config
from moe.library import Album, Extra, LibItem, Track

__all__ = ["import_album"]

log = logging.getLogger("moe.import")


class Hooks:
    """Import core plugin hook specifications."""

    @staticmethod
    @moe.hookspec
    def import_candidates(album: Album) -> Album:  # type: ignore
        """Imports candidate albums from implemented sources based on the given album.

        This hook should be used to import metadata from an external source and return
        a new album with the new metadata. The candidate albums will then be processed
        in the :meth:`process_candidates` hook.

        Note:
            This hook runs within the :meth:`pre_add` hook.

        Args:
            album: Album being added to the library.

        Returns:
            A new album with imported metadata.
        """  # noqa: DAR202

    @staticmethod
    @moe.hookspec
    def process_candidates(old_album: Album, candidates: list[Album]):
        """Process the imported candidate albums.

        If you wish to save and apply any candidate album metadata, it should be applied
        onto the original album, ``old_album``.

        Ensure any potential conflicts with existing albums in the database are
        resolved.

        Args:
            old_album: Album being added to the library.
            candidates: New candidate albums with imported metadata.
        """


@moe.hookimpl
def add_hooks(pm: pluggy.manager.PluginManager):
    """Registers `import` core hookspecs to Moe."""
    from moe.plugins.moe_import.import_core import Hooks

    pm.add_hookspecs(Hooks)


@moe.hookimpl
def pre_add(item: LibItem):
    """Fixes album metadata via external sources prior to it being added to the lib."""
    if isinstance(item, Album):
        album = item
    elif isinstance(item, (Extra, Track)):
        album = item.album_obj
    else:
        raise NotImplementedError

    import_album(album)


def import_album(album: Album):
    """Imports album metadata for an album."""
    log.debug(f"Importing album metadata. [{album=!r}]")

    candidates = config.CONFIG.pm.hook.import_candidates(album=album)
    config.CONFIG.pm.hook.process_candidates(
        old_album=album,
        candidates=candidates,
    )

    log.debug(f"Imported album metadata. [{album=!r}]")



"""Removes music from the library.

Note:
    This plugin is enabled by default.
"""

import argparse
import logging

import moe
import moe.cli
from moe.plugins import remove as moe_rm
from moe.util.cli import cli_query, query_parser

__all__: list[str] = []

log = logging.getLogger("moe.cli.remove")


@moe.hookimpl
def add_command(cmd_parsers: argparse._SubParsersAction):
    """Adds the ``remove`` command to Moe's CLI."""
    rm_parser = cmd_parsers.add_parser(
        "remove",
        aliases=["rm"],
        description="Removes music from the library.",
        help="remove music from the library",
        parents=[query_parser],
    )
    rm_parser.set_defaults(func=_parse_args)


def _parse_args(args: argparse.Namespace):
    """Parses the given commandline arguments.

    Args:
        args: Commandline arguments to parse.

    Raises:
        SystemExit: Invalid query given, or no items to remove.
    """
    items = cli_query(args.query, query_type=args.query_type)

    for item in items:
        moe_rm.remove_item(item)



"""An Album in the database and any related logic."""

import datetime
import logging
from pathlib import Path, PurePath
from typing import TYPE_CHECKING, Any, Optional, TypeVar, cast

import pluggy
import sqlalchemy as sa
from sqlalchemy import JSON, Column, Date, Integer, String
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import relationship

import moe
from moe import config
from moe.library.lib_item import LibItem, LibraryError, PathType, SABase

if TYPE_CHECKING:
    from moe.library.extra import Extra
    from moe.library.track import Track

__all__ = ["Album", "AlbumError"]

log = logging.getLogger("moe.album")


class Hooks:
    """Album hook specifications."""

    @staticmethod
    @moe.hookspec
    def create_custom_album_fields() -> dict[str, Any]:  # type: ignore
        """Creates new custom fields for an Album.

        Returns:
            Dict of the field names to their default values or ``None`` for no default.

        Example:
            .. code:: python

                return {"my_new_field": "default value", "other_field": None}

            You can then access your new field as if it were a normal field::

                album.my_new_field = "awesome new value"

        Important:
            Your custom field should follow the same naming rules as any other python
            variable i.e. no spaces, starts with a letter, and consists solely of
            alpha-numeric and underscore characters.
        """  # noqa: DAR202

    @staticmethod
    @moe.hookspec
    def is_unique_album(album: "Album", other: "Album") -> bool:  # type: ignore
        """Add new conditions to determine whether two albums are unique.

        "Uniqueness" is meant in terms of whether the two albums should be considered
        duplicates in the library. These additional conditions will be applied inside a
        album's :meth:`is_unique` method.
        """


@moe.hookimpl
def add_hooks(pm: pluggy.manager.PluginManager):
    """Registers `album` hookspecs to Moe."""
    from moe.library.album import Hooks

    pm.add_hookspecs(Hooks)


class AlbumError(LibraryError):
    """Error performing some operation on an Album."""


# Album generic, used for typing classmethod
A = TypeVar("A", bound="Album")


class Album(LibItem, SABase):
    """An album is a collection of tracks and represents a specific album release.

    Albums also house any attributes that are shared by tracks e.g. albumartist.

    Attributes:
        artist (str): AKA albumartist.
        country (str): Country the album was released in (two character identifier).
        date (datetime.date): Album release date.
        disc_total (int): Number of discs in the album.
        extras (list[Extra]): Extra non-track files associated with the album.
        label (str): Album release label.
        media (str): Album release format (e.g. CD, Digital, etc.)
        original_date (datetime.date): Date of the original release of the album.
        path (pathlib.Path): Filesystem path of the album directory.
        title (str)
        tracks (list[Track]): Album's corresponding tracks.
        year (int): Album release year. Note, this field is read-only. Set ``date``
            instead.
    """

    __tablename__ = "album"

    _id: int = cast(int, Column(Integer, primary_key=True))
    artist: str = cast(str, Column(String, nullable=False))
    country: str = cast(str, Column(String, nullable=True))
    date: datetime.date = cast(datetime.date, Column(Date, nullable=False))
    disc_total: int = cast(int, Column(Integer, nullable=False, default=1))
    label: str = cast(str, Column(String, nullable=True))
    media: str = cast(str, Column(String, nullable=True))
    original_date: datetime.date = cast(datetime.date, Column(Date, nullable=True))
    path: Path = cast(Path, Column(PathType, nullable=False, unique=True))
    title: str = cast(str, Column(String, nullable=False))
    _custom_fields: dict[str, Any] = cast(
        dict[str, Any],
        Column(
            MutableDict.as_mutable(JSON(none_as_null=True)),
            default="{}",
            nullable=False,
        ),
    )

    tracks: list["Track"] = relationship(
        "Track",
        back_populates="album_obj",
        cascade="all, delete-orphan",
        collection_class=list,
    )
    extras: list["Extra"] = relationship(
        "Extra",
        back_populates="album_obj",
        cascade="all, delete-orphan",
        collection_class=list,
    )

    def __init__(
        self,
        path: Path,
        artist: str,
        title: str,
        date: datetime.date,
        disc_total=1,
        **kwargs,
    ):
        """Creates an Album.

        Args:
            path: Filesystem path of the album directory.
            artist: Album artist.
            title: Album title.
            date: Album release date.
            disc_total: Number of discs in the album.
            **kwargs: Any other fields to assign to the album.
        """
        self._custom_fields = self._get_default_custom_fields()
        self._custom_fields_set = set(self._custom_fields)

        self.path = path
        self.artist = artist
        self.title = title
        self.date = date
        self.disc_total = disc_total

        for key, value in kwargs.items():
            setattr(self, key, value)

        if config.CONFIG.settings.original_date and self.original_date:
            self.date = self.original_date

        log.debug(f"Album created. [album={self!r}]")

    @classmethod
    def from_dir(cls: type[A], album_path: Path) -> A:
        """Creates an album from a directory.

        Args:
            album_path: Album directory path. The directory will be scanned for any
                files to be added to the album. Any non-track files will be added as
                extras.

        Returns:
            Created album.

        Raises:
            AlbumError: No tracks found in the given directory.
        """
        from moe.library.extra import Extra
        from moe.library.track import Track, TrackError

        log.debug(f"Creating album from directory. [dir={album_path}]")

        extra_paths = []
        album_file_paths = [path for path in album_path.rglob("*") if path.is_file()]
        album: Optional[Album] = None
        for file_path in album_file_paths:
            try:
                track = Track.from_file(file_path, album)
            except TrackError:
                extra_paths.append(file_path)
            else:
                if not album:
                    album = track.album_obj

        if not album:
            raise AlbumError(f"No tracks found in album directory. [dir={album_path}]")

        for extra_path in extra_paths:
            Extra(album, extra_path)

        log.debug(f"Album created from directory. [dir={album_path}, {album=!r}]")
        return album

    @property
    def fields(self) -> set[str]:
        """Returns any editable album fields."""
        return {
            "artist",
            "country",
            "date",
            "disc_total",
            "extras",
            "label",
            "media",
            "original_date",
            "path",
            "title",
            "tracks",
        }.union(self._custom_fields)

    def get_extra(self, rel_path: PurePath) -> Optional["Extra"]:
        """Gets an Extra by its path."""
        return next(
            (extra for extra in self.extras if extra.rel_path == rel_path), None
        )

    def get_track(self, track_num: int, disc: int = 1) -> Optional["Track"]:
        """Gets a Track by its track number."""
        return next(
            (
                track
                for track in self.tracks
                if track.track_num == track_num and track.disc == disc
            ),
            None,
        )

    def is_unique(self, other: "Album") -> bool:
        """Returns whether an album is unique in the library from ``other``."""
        if self.path == other.path:
            return False

        custom_uniqueness = config.CONFIG.pm.hook.is_unique_album(
            album=self, other=other
        )
        if False in custom_uniqueness:
            return False

        return True

    def merge(self, other: "Album", overwrite: bool = False) -> None:
        """Merges another album into this one.

        Args:
            other: Other album to be merged with the current album.
            overwrite: Whether or not to overwrite self if a conflict exists.
        """
        log.debug(f"Merging albums. [album_a={self!r}, album_b={other!r}")

        new_tracks: list["Track"] = []
        for other_track in other.tracks:
            conflict_track = self.get_track(other_track.track_num, other_track.disc)
            if conflict_track:
                conflict_track.merge(other_track, overwrite)
            else:
                new_tracks.append(other_track)
        self.tracks.extend(new_tracks)

        new_extras: list["Extra"] = []
        for other_extra in other.extras:
            conflict_extra = self.get_extra(other_extra.rel_path)
            if conflict_extra:
                conflict_extra.merge(other_extra, overwrite)
            else:
                new_extras.append(other_extra)
        self.extras.extend(new_extras)

        omit_fields = {"extras", "tracks"}
        for field in self.fields - omit_fields:
            other_value = getattr(other, field)
            self_value = getattr(self, field)
            if other_value and (overwrite or (not overwrite and not self_value)):
                setattr(self, field, other_value)

        log.debug(
            f"Albums merged. [album_a={self!r}, album_b={other!r}, {overwrite=!r}]"
        )

    @hybrid_property
    def original_year(self) -> int:  # type: ignore
        """Gets an Album's year."""
        return self.original_date.year

    @original_year.expression  # type: ignore
    def original_year(cls):  # noqa: B902
        """Returns a year at the sql level."""
        return sa.extract("year", cls.original_date)

    @hybrid_property
    def year(self) -> int:  # type: ignore
        """Gets an Album's year."""
        return self.date.year

    @year.expression  # type: ignore
    def year(cls):  # noqa: B902
        """Returns a year at the sql level."""
        return sa.extract("year", cls.date)

    def __eq__(self, other) -> bool:
        """Compares Albums by their fields."""
        if not isinstance(other, Album):
            return False

        omit_fields = {"extras", "tracks"}
        for field in self.fields - omit_fields:
            if not hasattr(other, field) or (
                getattr(self, field) != getattr(other, field)
            ):
                return False

        return True

    def __lt__(self, other: "Album") -> bool:
        """Sort an album based on its title, then artist, then date."""
        if self.title == other.title:
            if self.artist == other.artist:
                return self.date < other.date

            return self.artist < other.artist

        return self.title < other.title

    def __repr__(self):
        """Represents an Album using its fields."""
        field_reprs = []
        omit_fields = {"tracks", "extras"}
        for field in self.fields - omit_fields:
            if hasattr(self, field):
                field_reprs.append(f"{field}={getattr(self, field)!r}")
        repr_str = "Album(" + ", ".join(field_reprs)

        custom_field_reprs = []
        for custom_field, value in self._custom_fields.items():
            custom_field_reprs.append(f"{custom_field}={value}")
        if custom_field_reprs:
            repr_str += ", custom_fields=[" + ", ".join(custom_field_reprs) + "]"

        track_reprs = []
        for track in sorted(self.tracks):
            track_reprs.append(f"{track.disc}.{track.track_num} - {track.title}")
        repr_str += ", tracks=[" + ", ".join(track_reprs) + "]"

        extra_reprs = []
        for extra in sorted(self.extras):
            extra_reprs.append(f"{extra.path.name}")
        repr_str += ", extras=[" + ", ".join(extra_reprs) + "]"

        repr_str += ")"
        return repr_str

    def __str__(self):
        """String representation of an Album."""
        return f"{self.artist} - {self.title} ({self.year})"

    def _get_default_custom_fields(self) -> dict[str, Any]:
        """Returns the default custom album fields."""
        return {
            field: default_val
            for plugin_fields in config.CONFIG.pm.hook.create_custom_album_fields()
            for field, default_val in plugin_fields.items()
        }



"""Provides functionality to query the library for albums, extras, and tracks."""

import logging
import re
import shlex
from pathlib import Path

import sqlalchemy as sa
import sqlalchemy.orm
import sqlalchemy.sql.elements

from moe.config import MoeSession
from moe.library import Album, Extra, LibItem, Track

__all__: list[str] = ["QueryError", "query"]

log = logging.getLogger("moe.query")


class QueryError(Exception):
    """Error querying an item from the library."""


HELP_STR = r"""
The query must be in the format 'field:value' where field is, by default, a track's
field to match and value is that field's value (case-insensitive). To match an album's
field or an extra's field, prepend the field with `a:` or `e:` respectively.
Internally, this 'field:value' pair is referred to as a single term.

By default, tracks will be returned by the query, but you can choose to return albums
by using the ``-a, --album`` option, or you can return extras using the ``-e, --extra``
option.

If you would like to specify a value with whitespace or multiple words, enclose the
term in quotes.

SQL LIKE query syntax is used for normal queries, which means
the '_'  and '%' characters have special meaning:
% - The percent sign represents zero, one, or multiple characters.
_ - The underscore represents a single character.

To match these special characters as normal, use '/' as an escape character.

The value can also be a regular expression. To enforce this, use two colons
e.g. 'field::value.*'

As a shortcut to matching all entries, use '*' as the term.

Finally, you can also specify any number of terms.
For example, to match all Wu-Tang Clan tracks that start with the letter 'A', use:
'"artist:wu-tang clan" title:a%'

Note that when using multiple terms, they are joined together using AND logic, meaning
all terms must be true to return a match.

For more information on queries, see the docs.
https://mrmoe.readthedocs.io/en/latest/query.html
"""

# each query will be split into these groups
FIELD_TYPE = "field_type"
FIELD = "field"
SEPARATOR = "separator"
VALUE = "value"


def query(query_str: str, query_type: str) -> list[LibItem]:
    """Queries the database for items matching the given query string.

    Args:
        query_str: Query string to parse. See HELP_STR for more info.
        query_type: Type of library item to return: either 'album', 'extra', or 'track'.

    Returns:
        All items matching the query of type ``query_type``.

    Raises:
        QueryError: Invalid query.

    See Also:
        `The query docs <https://mrmoe.readthedocs.io/en/latest/query.html>`_
    """
    log.debug(f"Querying library for items. [{query_str=!r}, {query_type=!r}]")
    session = MoeSession()

    terms = shlex.split(query_str)
    if not terms:
        raise QueryError(f"No query given.\n{HELP_STR}")

    if query_type == "album":
        library_query = session.query(Album)
        if session.query(Track).first():
            library_query = library_query.join(Track)
    elif query_type == "extra":
        library_query = session.query(Extra).join(Album).join(Track)
    else:
        library_query = session.query(Track).join(Album)
    if query_type != "extra" and session.query(Extra).first():
        library_query = library_query.join(Extra)

    for term in terms:
        parsed_term = _parse_term(term)
        library_query = library_query.filter(
            _create_filter_expression(
                parsed_term[FIELD_TYPE],
                parsed_term[FIELD],
                parsed_term[SEPARATOR],
                parsed_term[VALUE],
            )
        )
    items = library_query.all()

    log.debug(f"Queried library for items. [{items=!r}]")
    return items


def _parse_term(term: str) -> dict[str, str]:
    """Parse the given database query term.

    A term is a single field:value declaration.

    Args:
        term: Query term to parse.

    Returns:
        A dictionary containing each named group and its value.
        The named groups are field_type, field, separator, and value.

    Example:
        >>> parse_term('a:artist:name')
        {"field_type": "album", "field": "artist", "separator": ":", "value": "name"}

    Note:
        * The fields are meant to be programatically accessed with the respective
        group constant e.g. `expression[FIELD] == "artist"`
        * All `field` values are automatically converted to lowercase.

    Raises:
        QueryError: Invalid query term.
    """
    # '*' is used as a shortcut to return all entries.
    # We use track_num as all tracks are guaranteed to have a track number.
    # '%' is an SQL LIKE query special character.
    if term == "*":
        return {FIELD_TYPE: "album", FIELD: "_id", SEPARATOR: ":", VALUE: "%"}

    query_re = re.compile(
        rf"""
        (?P<{FIELD_TYPE}>[aet]:)?
        (?P<{FIELD}>\S+?)
        (?P<{SEPARATOR}>::?)
        (?P<{VALUE}>.*)
        """,
        re.VERBOSE,
    )

    match = re.match(query_re, term)
    if not match:
        raise QueryError(f"Invalid query term. [{term=!r}]\n{HELP_STR}")

    match_dict = match.groupdict()
    match_dict[FIELD] = match_dict[FIELD].lower()
    if match_dict[FIELD_TYPE] == "a:":
        match_dict[FIELD_TYPE] = "album"
    elif match_dict[FIELD_TYPE] == "e:":
        match_dict[FIELD_TYPE] = "extra"
    else:
        match_dict[FIELD_TYPE] = "track"

    return match_dict


def _create_filter_expression(field_type: str, field: str, separator: str, value: str):
    """Maps a user-given query term to a filter expression for the database query.

    Args:
        field_type: LibItem type of ``field``.
        field: The field to query for.
        separator: Indicates the type of query.
        value: Value of ``field`` to match.

    Returns:
        A filter for the database query.

        A "filter" is anything accepted by a sqlalchemy `Query.filter()`.
        https://docs.sqlalchemy.org/en/13/orm/query.html#sqlalchemy.orm.query.Query.filter

    Raises:
        QueryError: Invalid query given.
    """
    attr = _get_field_attr(field, field_type)

    if separator == ":":
        if str(attr).endswith(".path"):
            return attr == Path(value)

        # normal string match query - should be case insensitive
        return attr.ilike(value, escape="/")

    elif separator == "::":
        try:
            re.compile(value)
        except re.error as re_err:
            raise QueryError(
                f"Invalid regular expression. [regex={value!r}]"
            ) from re_err

        return attr.op("regexp")(sa.sql.expression.literal(value))

    raise QueryError(f"Invalid query type separator. [{separator=!r}]")


def _get_field_attr(field: str, field_type: str):
    """Gets the corresponding attribute for the given field to use in a query filter."""
    if field == "genre":
        field = "genres"

    if field_type == "album":
        try:
            return getattr(Album, field)
        except AttributeError:
            # assume custom field
            custom_func = sa.func.json_each(
                Album._custom_fields, f"$.{field}"
            ).table_valued("value", joins_implicitly=True)
            return custom_func.c.value
    elif field_type == "extra":
        try:
            return getattr(Extra, field)
        except AttributeError:
            # assume custom field
            custom_func = sa.func.json_each(
                Extra._custom_fields, f"$.{field}"
            ).table_valued("value", joins_implicitly=True)
            return custom_func.c.value
    else:
        try:
            return getattr(Track, field)
        except AttributeError:
            # assume custom field
            custom_func = sa.func.json_each(
                Track._custom_fields, f"$.{field}"
            ).table_valued("value", joins_implicitly=True)
            return custom_func.c.value



"""Syncs library metadata with external sources."""

import moe
from moe import config

from . import sync_cli, sync_core
from .sync_core import *

__all__ = []
__all__.extend(sync_core.__all__)


@moe.hookimpl
def plugin_registration():
    """Only register the cli sub-plugin if the cli is enabled."""
    config.CONFIG.pm.register(sync_core, "sync_core")
    if config.CONFIG.pm.has_plugin("cli"):
        config.CONFIG.pm.register(sync_cli, "sync_cli")



"""Handles duplicate detection and resolution in the library."""

import logging

import moe
from moe import config

from . import dup_cli, dup_core
from .dup_core import *

__all__ = []
__all__.extend(dup_core.__all__)

log = logging.getLogger("moe.dup")


@moe.hookimpl
def plugin_registration():
    """Only register the cli sub-plugin if the cli is enabled."""
    config.CONFIG.pm.register(dup_core, "dup_core")
    if config.CONFIG.pm.has_plugin("cli"):
        config.CONFIG.pm.register(dup_cli, "dup_cli")



"""Prints information about music in the library.

All fields and their values should be printed to stdout for any music queried.

Note:
    This plugin is enabled by default.
"""

import argparse
import logging
from collections import OrderedDict
from typing import Any

import moe
import moe.cli
from moe import config
from moe.library import Album, Extra, LibItem, Track
from moe.util.cli import cli_query, query_parser

__all__: list[str] = []

log = logging.getLogger("moe.cli.info")


@moe.hookimpl
def plugin_registration():
    """Depend on the cli plugin."""
    if not config.CONFIG.pm.has_plugin("cli"):
        config.CONFIG.pm.set_blocked("info")
        log.warning("The 'info' plugin requires the 'cli' plugin to be enabled.")


@moe.hookimpl
def add_command(cmd_parsers: argparse._SubParsersAction):
    """Adds the ``info`` command to Moe's CLI."""
    info_parser = cmd_parsers.add_parser(
        "info",
        description="Prints information about music in the library.",
        help="print info for music in the library",
        parents=[query_parser],
    )
    info_parser.set_defaults(func=_parse_args)


def _parse_args(args: argparse.Namespace):
    """Parses the given commandline arguments.

    Args:
        args: Commandline arguments to parse.

    Raises:
        SystemExit: Invalid query given, or no items found.
    """
    items = cli_query(args.query, query_type=args.query_type)

    print(_fmt_infos(items), end="")


def _fmt_infos(items: list[LibItem]):
    """Formats information for multiple items together."""
    out_str = ""
    for item in items:
        out_str += _fmt_info(item) + "\n"

        if item is not items[-1]:
            out_str += "\n"

    return out_str


def _fmt_info(item: LibItem) -> str:
    """Formats the attribute/value pairs of an item into a str."""
    if isinstance(item, Track):
        return _fmt_track_info(item)
    elif isinstance(item, Album):
        return _fmt_album_info(item)
    elif isinstance(item, Extra):
        return _fmt_extra_info(item)

    raise NotImplementedError


def _fmt_album_info(album: Album) -> str:
    """Formats an album's information for display."""
    base_dict = _get_base_dict(album)
    base_dict.pop("extras", None)
    base_dict.pop("tracks", None)
    base_info = "\n".join(
        f"{field}: {value}"
        for field, value in OrderedDict(sorted(base_dict.items())).items()
    )

    tracks = "\nTracks\n"
    tracks += "\n".join(track.title for track in sorted(album.tracks))

    extras = "\nExtras\n"
    extras += "\n".join(extra.path.name for extra in sorted(album.extras))

    return base_info + "\n" + tracks + "\n" + extras


def _fmt_extra_info(extra: Extra) -> str:
    """Formats a extra's information for display."""
    base_dict = _get_base_dict(extra)

    return "\n".join(
        f"{field}: {value}"
        for field, value in OrderedDict(sorted(base_dict.items())).items()
    )


def _fmt_track_info(track: Track) -> str:
    """Formats a track's information for display."""
    base_dict = OrderedDict(sorted(_get_base_dict(track).items()))
    base_dict.pop("album_obj", None)

    return "\n".join(
        f"{field}: {value}"
        for field, value in OrderedDict(sorted(base_dict.items())).items()
    )


def _get_base_dict(item: LibItem) -> dict[str, Any]:
    """Represents an item as a dictionary.

    Only public attributes that are not empty will be included. Also, any attributes
    that are not relevant to the user e.g. sqlalchemy specific attributes will
    be removed.

    Args:
        item: Library item.

    Returns:
        Returns a dict representation of an Item in the form { attribute: value }.
    """
    item_dict = {}
    for attr in sorted(item.fields):
        value = getattr(item, attr)
        if value:
            item_dict[attr] = value

    return item_dict



"""Adds the ``add`` command to add music to the library."""

import argparse
import logging
from pathlib import Path

import moe
import moe.cli
from moe.library import Album, AlbumError, Track, TrackError
from moe.plugins import add as moe_add
from moe.plugins.add.add_core import AddError
from moe.util.cli import PromptChoice

log = logging.getLogger("moe.cli.add")

__all__: list[str] = []


class SkipAdd(Exception):
    """Used to skip adding a single item."""


@moe.hookimpl
def add_command(cmd_parsers: argparse._SubParsersAction):
    """Adds the ``add`` command to Moe's CLI."""
    add_parser = cmd_parsers.add_parser(
        "add", description="Adds music to the library.", help="add music to the library"
    )
    add_parser.add_argument(
        "paths",
        metavar="path",
        nargs="+",
        help="dir to add an album or file to add a track",
    )
    add_parser.set_defaults(func=_parse_args)


@moe.hookimpl
def add_import_prompt_choice(prompt_choices: list[PromptChoice]):
    """Adds the ``skip`` prompt choice to the import prompt."""
    prompt_choices.append(
        PromptChoice(title="Skip", shortcut_key="s", func=_skip_import)
    )


def _skip_import(
    old_album: Album,
    new_album: Album,
):
    """Skip adding/importing the current item."""
    raise SkipAdd()


def _parse_args(args: argparse.Namespace):
    """Parses the given commandline arguments.

    Tracks can be added as files or albums as directories.

    Args:
        args: Commandline arguments to parse.

    Raises:
        SystemExit: Path given does not exist.
    """
    paths = [Path(arg_path) for arg_path in args.paths]

    error_count = 0
    for path in paths:
        try:
            _add_path(path)
        except (AddError, AlbumError, TrackError) as err:
            log.error(err)
            error_count += 1
        except SkipAdd:
            log.debug(f"Skipped adding item. [{path=}]")

    if error_count:
        raise SystemExit(1)


def _add_path(path: Path):
    """Adds an item to the library from a given path.

    Args:
        path: Path to add. Either a directory for an Album or a file for a Track.

    Raises:
        AddError: Path not found or other issue adding the item to the library.
        AlbumError: Could not create an album from the given directory.
        TrackError: Could not create a track from the given file.
        SkipAdd: External program or user elected to skip adding the item.
    """
    if path.is_file():
        moe_add.add_item(Track.from_file(path))
    elif path.is_dir():
        moe_add.add_item(Album.from_dir(path))
    else:
        log.error(f"Path not found. [{path=}]")
        raise AddError("Path not found. [{path=}]")



"""Functionality related to prompting the user for some action."""

import logging
import operator
from dataclasses import dataclass
from typing import Callable

import questionary

__all__ = ["PromptChoice", "choice_prompt"]

log = logging.getLogger("moe.cli")


@dataclass
class PromptChoice:
    """A single, user-selectable choice for a CLI prompt.

    Attributes:
        title: Title of the prompt choice that is displayed to the user.
        shortcut_key: Single character the user will use to select the choice.

            Important:
                Ensure each shortcut key is not in use by another PromptChoice.
        func: The function that should get called if a choice is selected.
            The definition for how to call ``func`` should be specified by the plugin.
    """

    title: str
    shortcut_key: str
    func: Callable


def choice_prompt(
    prompt_choices: list[PromptChoice], question: str = "What do you want to do?"
) -> PromptChoice:
    """Generates a user choice prompt.

    Args:
        prompt_choices: Prompt choices to be used.
        question: Question prompted to the user.

    Returns:
        The chosen prompt choice.

    Raises:
        SystemExit: Invalid user input.
    """
    prompt_choices.sort(key=operator.attrgetter("shortcut_key"))

    questionary_choices: list[questionary.Choice] = []
    for prompt_choice in prompt_choices:
        questionary_choices.append(
            questionary.Choice(
                title=prompt_choice.title,
                shortcut_key=prompt_choice.shortcut_key,
                value=prompt_choice.shortcut_key,
            )
        )

    user_input = questionary.rawselect(question, choices=questionary_choices).ask()

    for prompt_choice in prompt_choices:
        if prompt_choice.shortcut_key == user_input:
            return prompt_choice

    log.error(f"Invalid option selected. [{user_input=!r}]")
    raise SystemExit(1)



"""Adds the ``sync`` command to sync library metadata."""

import argparse
import logging

import moe
import moe.cli
from moe.plugins import sync as moe_sync
from moe.util.cli import cli_query, query_parser

log = logging.getLogger("moe.cli.sync")

__all__: list[str] = []


@moe.hookimpl
def add_command(cmd_parsers: argparse._SubParsersAction):
    """Adds the ``add`` command to Moe's CLI."""
    add_parser = cmd_parsers.add_parser(
        "sync",
        description="Sync music metadata.",
        help="sync music metadata",
        parents=[query_parser],
    )
    add_parser.set_defaults(func=_parse_args)


def _parse_args(args: argparse.Namespace):
    """Parses the given commandline arguments.

    Args:
        args: Commandline arguments to parse.

    Raises:
        SystemExit: Invalid query given or query returned no items.
    """
    items = cli_query(args.query, query_type=args.query_type)

    for item in items:
        moe_sync.sync_item(item)



"""Imports metadata for music in your library."""

import sys

import moe
from moe import config

from . import import_cli, import_core
from .import_cli import *
from .import_core import *

__all__ = []
__all__.extend(import_cli.__all__)
__all__.extend(import_core.__all__)


@moe.hookimpl
def plugin_registration():
    """Only register the cli sub-plugin if the cli is enabled."""
    config.CONFIG.pm.register(import_core, "import_core")
    if config.CONFIG.pm.has_plugin("cli"):
        config.CONFIG.pm.register(import_cli, "import_cli")

    # re-register under the "import" name instead of "moe_import"
    config.CONFIG.pm.unregister(name="moe_import")
    config.CONFIG.pm.register(sys.modules[__name__], "import")



"""new field: media.

Revision ID: 60608d9d2908
Revises: fe0956c93730
Create Date: 2022-10-06 20:26:02.640360

"""
import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision = "60608d9d2908"
down_revision = "fe0956c93730"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("album", schema=None) as batch_op:
        batch_op.add_column(sa.Column("media", sa.String(), nullable=True))


def downgrade():
    with op.batch_alter_table("album", schema=None) as batch_op:
        batch_op.drop_column("media")



"""Writes tags to track files."""

import logging

import mediafile
import pluggy

import moe
from moe import config
from moe.library import LibItem, Track

__all__ = ["write_tags"]

log = logging.getLogger("moe.write")


class Hooks:
    """Write plugin hook specifications."""

    @staticmethod
    @moe.hookspec
    def write_custom_tags(track: Track):
        """Allow plugins to write tags to a Track.

        How you write tags to the track is up to each individual plugin. Internally,
        Moe uses the `mediafile <https://github.com/beetbox/mediafile>`_ library to
        write tags.

        Args:
            track: Track to write tags to.

        Example:
            .. code:: python

                audio_file = mediafile.MediaFile(track.path)
                audio_file.album = track.album
                audio_file.save()

        See Also:
            * :ref:`Album and track fields <fields:Fields>`
            * `Mediafile docs <https://mediafile.readthedocs.io/en/latest/>`_
            * The :meth:`~moe.library.track.Hooks.read_custom_tags` hook for reading
              tags.
        """


@moe.hookimpl
def add_hooks(pm: pluggy.manager.PluginManager):
    """Registers `write` hookspecs to Moe."""
    from moe.plugins.write import Hooks

    pm.add_hookspecs(Hooks)


@moe.hookimpl
def process_new_items(items: list[LibItem]):
    """Writes tags to any new tracks in the library."""
    for item in items:
        if isinstance(item, Track):
            write_tags(item)


@moe.hookimpl
def process_changed_items(items: list[LibItem]):
    """Writes tags to any altered tracks in the library."""
    for item in items:
        if isinstance(item, Track):
            write_tags(item)


@moe.hookimpl(tryfirst=True)
def write_custom_tags(track: Track):
    """Writes all internally tracked tags to the track."""
    audio_file = mediafile.MediaFile(track.path)

    audio_file.album = track.album
    audio_file.albumartist = track.albumartist
    audio_file.artist = track.artist
    audio_file.artists = track.artists
    audio_file.country = track.album_obj.country
    audio_file.date = track.album_obj.date
    audio_file.disc = track.disc
    audio_file.disctotal = track.album_obj.disc_total
    audio_file.genres = track.genres
    audio_file.label = track.album_obj.label
    audio_file.media = track.album_obj.media
    audio_file.original_date = track.album_obj.original_date
    audio_file.title = track.title
    audio_file.track = track.track_num

    audio_file.save()


def write_tags(track: Track):
    """Write tags to a track's file."""
    log.debug(f"Writing tags to track. [{track=!r}]")

    config.CONFIG.pm.hook.write_custom_tags(track=track)

    log.info(f"Wrote tags to track. [{track=!r}]")



"""Lists music in the library.

Note:
    This plugin is enabled by default.
"""

import argparse
import logging

import moe
import moe.cli
from moe import config
from moe.util.cli import cli_query, query_parser

__all__: list[str] = []

log = logging.getLogger("moe.cli.list")


@moe.hookimpl
def plugin_registration():
    """Depend on the cli plugin."""
    if not config.CONFIG.pm.has_plugin("cli"):
        config.CONFIG.pm.set_blocked("list")
        log.warning("The 'list' plugin requires the 'cli' plugin to be enabled.")


@moe.hookimpl
def add_command(cmd_parsers: argparse._SubParsersAction):
    """Adds the ``list`` command to Moe's CLI."""
    ls_parser = cmd_parsers.add_parser(
        "list",
        aliases=["ls"],
        description="Lists music in the library.",
        help="list music in the library",
        parents=[query_parser],
    )
    ls_parser.add_argument(
        "-p",
        "--paths",
        action="store_true",
        help="list paths",
    )
    ls_parser.set_defaults(func=_parse_args)


def _parse_args(args: argparse.Namespace):
    """Parses the given commandline arguments.

    Args:
        args: Commandline arguments to parse.

    Raises:
        SystemExit: Invalid query or no items found.
    """
    items = cli_query(args.query, query_type=args.query_type)

    for item in sorted(items):
        if args.paths:
            print(item.path)
        else:
            print(item)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['moe',
 'moe.library',
 'moe.plugins',
 'moe.plugins.add',
 'moe.plugins.duplicate',
 'moe.plugins.edit',
 'moe.plugins.moe_import',
 'moe.plugins.move',
 'moe.plugins.musicbrainz',
 'moe.plugins.remove',
 'moe.plugins.sync',
 'moe.util',
 'moe.util.cli',
 'moe.util.core']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.4.15,<2.0.0',
 'Unidecode>=1.2.0,<2.0.0',
 'alembic>=1.4.2,<2.0.0',
 'dynaconf>=3.1.4,<4.0.0',
 'mediafile>=0.9.0,<0.10.0',
 'musicbrainzngs>=0.7.1,<0.8.0',
 'pluggy>=0.13.1,<0.14.0',
 'pyyaml>=5.3.1,<6.0.0',
 'questionary>=1.9.0,<2.0.0',
 'rich>=12.5.1,<13.0.0']

extras_require = \
{'docs': ['furo>=2021.7.5-beta.38,<2022.0.0', 'Sphinx>=4.0.2,<5.0.0']}

entry_points = \
{'console_scripts': ['moe = moe.cli:main']}

setup_kwargs = {
    'name': 'moe',
    'version': '0.15.1',
    'description': 'The ultimate tool for managing your music library.',
    'long_description': "###############\nWelcome to Moe!\n###############\nMoe is our resident Music-Organizer-Extraordinaire who's sole purpose is to give you full control over your music library. In other words, it's a commandline-interface for managing your music.\n\n*******************\nDevelopment Warning\n*******************\nMoe is currently in early development, and is still training for his goal to become your all-powerful music library assistant. I do not recommend using Moe for anything except experimentation or testing until v1.0 is released.\n\nSo what can Moe do right now?\n=============================\n* Add music to your library, fixing tags with metadata from Musicbrainz.\n* Organize, remove, list, and edit your music in the library.\n* Supports including extra files with an album e.g. log or playlist files.\n* Supports tags with multiple values.\n\nIf you want to learn more, check out the `Getting Started <https://mrmoe.readthedocs.io/en/latest/getting_started.html>`_ docs.\n",
    'author': 'Jacob Pavlock',
    'author_email': 'jtpavlock@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/MoeMusic/Moe',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



"""Core musicbrainz api.

The ``musicbrainz`` core plugin provides the following functionality:

* Imports metadata from musicbrainz when adding a track or album to the library.
* Optionally updates a musicbrainz collection when items are added or removed from the
  library.

Note:
    This plugin is enabled by default.

See Also:
    * https://musicbrainz.org/doc/MusicBrainz_API/
    * https://python-musicbrainzngs.readthedocs.io/en/latest/api/
"""

import datetime
import logging
from pathlib import Path
from typing import Any, Callable, Optional

import dynaconf
import mediafile
import musicbrainzngs
import pkg_resources

import moe
from moe import config
from moe.library import Album, LibItem, Track

__all__ = [
    "MBAuthError",
    "add_releases_to_collection",
    "get_album_by_id",
    "get_matching_album",
    "get_track_by_id",
    "rm_releases_from_collection",
    "set_collection",
    "update_album",
]

log = logging.getLogger("moe.mb")

MAX_SEARCH_LIMIT = 100
"""Max number of entries that can be returned from a search to musicbrainz. This limit
is enforced by their api."""


class MBAuthError(Exception):
    """Musicbrainz user authentication error."""


musicbrainzngs.set_useragent(
    "moe",
    pkg_resources.get_distribution("moe").version,
    contact="https://mrmoe.readthedocs.io/en/latest/index.html",
)

# information to include in the release api query
RELEASE_INCLUDES = [
    "aliases",
    "artists",
    "artist-credits",
    "artist-rels",
    "labels",
    "media",
    "recordings",
    "recording-level-rels",
    "release-groups",
    "work-level-rels",
    "work-rels",
]


@moe.hookimpl
def add_config_validator(settings: dynaconf.base.LazySettings):
    """Validates musicbrainz plugin configuration settings."""
    login_required = False

    settings.validators.register(
        dynaconf.Validator(
            "musicbrainz.collection.auto_add",
            "musicbrainz.collection.auto_remove",
            default=False,
        )
    )

    if settings.get("musicbrainz.collection.auto_add", False) or settings.get(
        "musicbrainz.collection.auto_remove", False
    ):
        login_required = True
        settings.validators.register(
            dynaconf.Validator("musicbrainz.collection.collection_id", must_exist=True)
        )

    if login_required:
        settings.validators.register(
            dynaconf.Validator(
                "musicbrainz.username", "musicbrainz.password", must_exist=True
            )
        )


@moe.hookimpl
def create_custom_album_fields() -> dict[str, Any]:  # type: ignore
    """Adds relevant musicbrainz fields to an album."""
    return {"mb_album_id": None}


@moe.hookimpl
def create_custom_track_fields() -> dict[str, Any]:  # type: ignore
    """Adds relevant musicbrainz fields to a track."""
    return {"mb_track_id": None}


@moe.hookimpl
def import_candidates(album: Album) -> Album:
    """Applies musicbrainz metadata changes to a given album.

    Args:
        album: Original album used to search musicbrainz for a matching album.

    Returns:
        A new album containing all the corrected metadata from musicbrainz. Note, this
        album is not complete, as it will not contain any references to the filesystem.
    """
    return get_matching_album(album)


@moe.hookimpl
def process_removed_items(items: list[LibItem]):
    """Removes a release from a collection when removed from the library."""
    if not config.CONFIG.settings.musicbrainz.collection.auto_remove:
        return

    mb_ids = []
    for item in items:
        if isinstance(item, Album) and item.mb_album_id:
            mb_ids.append(item.mb_album_id)

    if mb_ids:
        try:
            rm_releases_from_collection(set(mb_ids))
        except MBAuthError as err:
            log.error(err)


@moe.hookimpl
def process_new_items(items: list[LibItem]):
    """Updates a user collection in musicbrainz with new releases."""
    if not config.CONFIG.settings.musicbrainz.collection.auto_add:
        return

    releases = set()
    for item in items:
        if isinstance(item, Album) and item.mb_album_id:
            releases.add(item.mb_album_id)

    if releases:
        try:
            add_releases_to_collection(releases)
        except MBAuthError as err:
            log.error(err)


@moe.hookimpl
def read_custom_tags(
    track_path: Path, album_fields: dict[str, Any], track_fields: dict[str, Any]
) -> None:
    """Read and set musicbrainz release IDs from a track file."""
    audio_file = mediafile.MediaFile(track_path)

    album_fields["mb_album_id"] = audio_file.mb_albumid
    track_fields["mb_track_id"] = audio_file.mb_releasetrackid


@moe.hookimpl
def sync_metadata(item: LibItem):
    """Sync musibrainz metadata for associated items."""
    if isinstance(item, Album) and hasattr(item, "mb_album_id"):
        item.merge(get_album_by_id(item.mb_album_id), overwrite=True)
    elif isinstance(item, Track) and hasattr(item, "mb_track_id"):
        item.merge(
            get_track_by_id(item.mb_track_id, item.album_obj.mb_album_id),
            overwrite=True,
        )


@moe.hookimpl
def write_custom_tags(track: Track):
    """Write musicbrainz ID fields as tags."""
    audio_file = mediafile.MediaFile(track.path)

    audio_file.mb_albumid = track.album_obj.mb_album_id
    audio_file.mb_releasetrackid = track.mb_track_id

    audio_file.save()


def add_releases_to_collection(
    releases: set[str], collection: Optional[str] = None
) -> None:
    """Adds releases to a musicbrainz collection.

    Args:
        releases: Musicbrainz release IDs to add to the collection.
        collection: Musicbrainz collection ID to add the releases to.
            If not given, defaults to the ``musicbrainz.collection.collection_id``
            config option.

    Raises:
        MBAuthError: Invalid musicbrainz user credentials in the configuration.
    """
    collection = (
        collection or config.CONFIG.settings.musicbrainz.collection.collection_id
    )

    log.debug(
        f"Adding releases to musicbrainz collection. [{releases=!r}, {collection=!r}]"
    )

    _mb_auth_call(
        musicbrainzngs.add_releases_to_collection,
        collection=collection,
        releases=releases,
    )

    log.info(
        f"Added releases to musicbrainz collection. [{releases=!r}, {collection=!r}]"
    )


def rm_releases_from_collection(
    releases: set[str], collection: Optional[str] = None
) -> None:
    """Removes releases from a musicbrainz collection.

    Args:
        releases: Musicbrainz release IDs to remove from the collection.
        collection: Musicbrainz collection ID to remove the releases from.
            If not given, defaults to the ``musicbrainz.collection.collection_id``
            config option.

    Raises:
        MBAuthError: Invalid musicbrainz user credentials in the configuration.
    """
    collection = (
        collection or config.CONFIG.settings.musicbrainz.collection.collection_id
    )

    log.debug(
        "Removing releases from musicbrainz collection. "
        f"[{releases=!r}, {collection=!r}]"
    )

    _mb_auth_call(
        musicbrainzngs.remove_releases_from_collection,
        collection=collection,
        releases=releases,
    )

    log.info(
        "Removed releases from musicbrainz collection. "
        f"[{releases=!r}, {collection=!r}]"
    )


def _mb_auth_call(api_func: Callable, **kwargs) -> Any:
    """Call a musicbrainz API function that requires user authentication.

    Args:
        api_func: Musicbrainz API function to call.
        **kwargs: Keyword arguments to pass to the API function call.

    Returns:
        The return value of the API function called.

    Raises:
        MBAuthError: Invalid user credentials in the configuration.
    """
    musicbrainzngs.auth(
        config.CONFIG.settings.musicbrainz.username,
        config.CONFIG.settings.musicbrainz.password,
    )

    try:
        return api_func(**kwargs)
    except musicbrainzngs.AuthenticationError as err:
        raise MBAuthError("User authentication with musicbrainz failed.") from err


def set_collection(releases: set[str], collection: Optional[str] = None) -> None:
    """Sets a musicbrainz collection with the given releases.

    The releases in the collection will be set to ``releases``, adding any releases not
    present in the collection, as well as removing any extraneous releases.

    Args:
        releases: Musicbrainz releases to set the collection to.
        collection: Musicbrainz collection ID for the collection to set.
            If not given, defaults to the ``musicbrainz.collection.collection_id``
            config option.

    Raises:
        MBAuthError: Invalid user credentials in the configuration.
    """
    collection = (
        collection or config.CONFIG.settings.musicbrainz.collection.collection_id
    )

    log.debug("Setting musicbrainz collection. " f"[{releases=!r}, {collection=!r}]")

    current_releases = []
    num_searches = 0
    limit = MAX_SEARCH_LIMIT
    while len(current_releases) == limit * num_searches:  # hitting the search limit
        result = _mb_auth_call(
            musicbrainzngs.get_releases_in_collection,
            collection=collection,
            limit=limit,
            offset=limit * num_searches,
        )
        for release in result["collection"]["release-list"]:
            current_releases.append(release["id"])

        num_searches += 1
    current_releases = set(current_releases)

    stale_releases = current_releases.difference(releases)
    rm_releases_from_collection(stale_releases, collection)

    new_releases = releases.difference(current_releases)
    add_releases_to_collection(new_releases, collection)


def get_matching_album(album: Album) -> Album:
    """Gets a matching musicbrainz album for a given album.

    Args:
        album: Album used to search for the release.

    Returns:
        An Album containing all musicbrainz metadata.
    """
    if album.mb_album_id:
        return get_album_by_id(album.mb_album_id)

    log.debug(f"Determing matching releases from musicbrainz. [{album=!r}]")

    search_criteria: dict = {}
    search_criteria["artist"] = album.artist
    search_criteria["release"] = album.title
    search_criteria["date"] = album.date.isoformat()

    releases = musicbrainzngs.search_releases(limit=1, **search_criteria)

    release = releases["release-list"][0]
    release_id = release["id"]

    log.info(f"Determined matching release from musicbrainz. [match={release_id!r}]")

    return get_album_by_id(release_id)  # searching by id provides more info


def get_album_by_id(release_id: str) -> Album:
    """Gets a musicbrainz album from a release ID.

    Args:
        release_id: Musicbrainz release ID to search.

    Returns:
        An album containing all metadata from the given ``release_id``.
    """
    log.debug(f"Fetching release from musicbrainz. [release={release_id!r}]")

    release = musicbrainzngs.get_release_by_id(release_id, includes=RELEASE_INCLUDES)

    log.info(f"Fetched release from musicbrainz. [release={release_id!r}]")

    return _create_album(release["release"])


def _create_album(release: dict) -> Album:
    """Creates an album from a given musicbrainz release."""
    log.debug(f"Creating album from musicbrainz release. [release={release['id']!r}]")

    album = Album(
        artist=_flatten_artist_credit(release["artist-credit"]),
        country=release["country"],
        date=_parse_date(release["date"]),
        disc_total=int(release["medium-count"]),
        label=release["label-info-list"][0]["label"]["name"],
        mb_album_id=release["id"],
        media=release["medium-list"][0]["format"],
        original_date=_parse_date(release["release-group"]["first-release-date"]),
        title=release["title"],
        path=None,  # type: ignore # this will get set in `add_prompt`
    )
    for medium in release["medium-list"]:
        for track in medium["track-list"]:
            Track(
                album=album,
                track_num=int(track["position"]),
                path=None,  # type: ignore # this will get set in `add_prompt`
                artist=_flatten_artist_credit(track["recording"]["artist-credit"]),
                disc=int(medium["position"]),
                mb_track_id=track["id"],
                title=track["recording"]["title"],
            )

    log.debug(f"Created album from musicbrainz release. [{album=!r}]")
    return album


def _flatten_artist_credit(artist_credit: list[dict]) -> str:
    """Given a musicbrainz formatted artist-credit, return the full artist name."""
    full_artist = ""
    for artist in artist_credit:
        if isinstance(artist, str):
            full_artist += artist
        else:
            full_artist += artist["artist"]["name"]

    return full_artist


def _parse_date(date: str) -> datetime.date:
    """Parses a date from a musicbrainz release."""
    date_parts = date.split("-")

    year = int(date_parts[0])
    try:
        month = int(date_parts[1])
    except IndexError:
        month = 1
    try:
        day = int(date_parts[2])
    except IndexError:
        day = 1

    return datetime.date(year, month, day)


def get_track_by_id(track_id: str, album_id: str) -> Track:
    """Gets a musicbrainz track from a given track and release id.

    Args:
        track_id: Musicbrainz track ID to match.
        album_id: Release album ID the track belongs to.

    Returns:
        A track containing all metadata associated with the given IDs.

    Raises:
        ValueError: Couldn't find track based on given ``track_id`` and ``album_id``.
    """
    log.debug(f"Fetching track from musicbrainz. [{track_id=!r}, {album_id=!r}]")

    album = get_album_by_id(album_id)
    for track in album.tracks:
        if track.mb_track_id == track_id:
            log.info(f"Fetched track from musicbrainz. [{track=!r}]")
            return track

    raise ValueError(
        f"Given track or album id could not be found. [{track_id=!r}, {album_id=!r}]"
    )


def update_album(album: Album):
    """Updates an album with metadata from musicbrainz.

    Args:
        album: Album to update.

    Raises:
        ValueError: ``album`` has no ``mb_album_id``.
    """
    log.debug(f"Updating album with musicbrainz metadata. [{album=!r}]")

    if not album.mb_album_id:
        raise ValueError(
            "Unable to update album, no musicbrainz id found. [{album=!r}]"
        )

    album.merge(get_album_by_id(album.mb_album_id), overwrite=True)



"""new field: artists.

Revision ID: fe0956c93730
Revises: 6d4e785df5cb
Create Date: 2022-10-06 18:26:11.939381

"""
import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision = "fe0956c93730"
down_revision = "6d4e785df5cb"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "artist",
        sa.Column("_id", sa.Integer(), nullable=False),
        sa.Column("_track_id", sa.Integer(), nullable=True),
        sa.Column("name", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["_track_id"],
            ["track._id"],
        ),
        sa.PrimaryKeyConstraint("_id"),
    )


def downgrade():
    op.drop_table("artist")



"""Shared functionality between library albums, extras, and tracks."""

import logging
from pathlib import Path
from typing import Any

import pluggy
import sqlalchemy
import sqlalchemy as sa
import sqlalchemy.event
import sqlalchemy.orm
from sqlalchemy.orm import declarative_base

import moe
from moe import config

__all__ = ["LibItem", "LibraryError"]

log = logging.getLogger("moe.lib_item")

SABase = declarative_base()


class LibraryError(Exception):
    """General library error."""


class Hooks:
    """General usage library item hooks."""

    @staticmethod
    @moe.hookspec
    def edit_changed_items(items: list["LibItem"]):
        """Edit items in the library that were changed in some way.

        Args:
            items: Any changed items that existed in the library prior to the current
                session.

        See Also:
            The :meth:`process_changed_items` hook for processing items with finalized
            values.
        """

    @staticmethod
    @moe.hookspec
    def edit_new_items(items: list["LibItem"]):
        """Edit new items in the library.

        Args:
            items: Any items being added to the library for the first time.

        See Also:
            The :meth:`process_new_items` hook for processing items with finalized
            values.
        """

    @staticmethod
    @moe.hookspec
    def process_removed_items(items: list["LibItem"]):
        """Process items that have been removed from the library.

        Args:
            items: Any items that existed in the library prior to the current session,
                but have now been removed from the library.
        """

    @staticmethod
    @moe.hookspec
    def process_changed_items(items: list["LibItem"]):
        """Process items in the library that were changed in some way.

        Args:
            items: Any changed items that existed in the library prior to the current
                session.

        Important:
            Any changes made to the items will be lost.

        See Also:
            The :meth:`edit_changed_items` hook for editing items before their values
            are finalized.
        """

    @staticmethod
    @moe.hookspec
    def process_new_items(items: list["LibItem"]):
        """Process new items in the library.

        Args:
            items: Any items being added to the library for the first time.

        Important:
            Any changes made to the items will be lost.

        See Also:
            The :meth:`edit_new_items` hook for editing items before their values are
            finalized.
        """


@moe.hookimpl
def add_hooks(pm: pluggy.manager.PluginManager):
    """Registers `add` hookspecs to Moe."""
    from moe.library.lib_item import Hooks

    pm.add_hookspecs(Hooks)


@moe.hookimpl
def register_sa_event_listeners(session: sqlalchemy.orm.Session):
    """Registers event listeners for editing and processing new items."""
    sqlalchemy.event.listen(
        session,
        "before_flush",
        _edit_before_flush,
    )
    sqlalchemy.event.listen(
        session,
        "after_flush",
        _process_after_flush,
    )


def _edit_before_flush(
    session: sqlalchemy.orm.Session,
    flush_context: sqlalchemy.orm.UOWTransaction,
    instances,
):
    """Runs the ``edit_*_items`` hook specifications before items are flushed.

    This uses the sqlalchemy ORM event ``before_flush`` in the background to determine
    the time of execution and to provide any new, changed, or deleted items to the hook
    implementations.

    Args:
        session: Current db session.
        flush_context: sqlalchemy obj which handles the details of the flush.
        instances: Objects passed to the `session.flush()` method (deprecated).

    See Also:
        `SQLAlchemy docs on state management <https://docs.sqlalchemy.org/en/14/orm/session_state_management.html>`_
    """  # noqa: E501
    changed_items = []
    for dirty_item in session.dirty:
        if session.is_modified(dirty_item) and isinstance(dirty_item, LibItem):
            changed_items.append(dirty_item)
    if changed_items:
        log.debug(f"Editing changed items. [{changed_items=!r}]")
        config.CONFIG.pm.hook.edit_changed_items(items=changed_items)
        log.debug(f"Edited changed items. [{changed_items=!r}]")

    new_items = []
    for new_item in session.new:
        if isinstance(new_item, LibItem):
            new_items.append(new_item)
    if new_items:
        log.debug(f"Editing new items. [{new_items=!r}]")
        config.CONFIG.pm.hook.edit_new_items(items=new_items)
        log.debug(f"Edited new items. [{new_items=!r}]")


def _process_after_flush(
    session: sqlalchemy.orm.Session,
    flush_context: sqlalchemy.orm.UOWTransaction,
):
    """Runs the ``process_*_items`` hook specifications after items are flushed.

    This uses the sqlalchemy ORM event ``after_flush`` in the background to determine
    the time of execution and to provide any new, changed, or deleted items to the hook
    implementations.

    Args:
        session: Current db session.
        flush_context: sqlalchemy obj which handles the details of the flush.

    See Also:
        `SQLAlchemy docs on state management <https://docs.sqlalchemy.org/en/14/orm/session_state_management.html>`_
    """  # noqa: E501
    changed_items = []
    for dirty_item in session.dirty:
        if session.is_modified(dirty_item) and isinstance(dirty_item, LibItem):
            changed_items.append(dirty_item)
    if changed_items:
        log.debug(f"Processing changed items. [{changed_items=!r}]")
        config.CONFIG.pm.hook.process_changed_items(items=changed_items)
        log.debug(f"Processed changed items. [{changed_items=!r}]")

    new_items = []
    for new_item in session.new:
        if isinstance(new_item, LibItem):
            new_items.append(new_item)
    if new_items:
        log.debug(f"Processing new items. [{new_items=!r}]")
        config.CONFIG.pm.hook.process_new_items(items=new_items)
        log.debug(f"Processed new items. [{new_items=!r}]")

    removed_items = []
    for removed_item in session.deleted:
        if isinstance(removed_item, LibItem):
            removed_items.append(removed_item)
    if removed_items:
        log.debug(f"Processing removed items. [{removed_items=!r}]")
        config.CONFIG.pm.hook.process_removed_items(items=removed_items)
        log.debug(f"Processed removed items. [{removed_items=!r}]")


class LibItem:
    """Base class for library items i.e. Albums, Extras, and Tracks."""

    _custom_fields_set = None

    @property
    def path(self) -> Path:
        """A library item's filesystem path."""
        raise NotImplementedError

    @property
    def custom_fields(self) -> set[str]:
        """Returns the custom fields of an item."""
        if self._custom_fields_set is None:
            object.__setattr__(
                self, "_custom_fields_set", set(self._get_default_custom_fields())
            )

        assert self._custom_fields_set is not None
        return self._custom_fields_set

    def _get_default_custom_fields(self) -> dict[str, Any]:
        """Returns the default custom fields of an item."""
        raise NotImplementedError

    @property
    def fields(self) -> set[str]:
        """Returns the editable fields of an item."""
        raise NotImplementedError

    def is_unique(self, other: "LibItem") -> bool:
        """Returns whether an item is unique in the library from ``other``."""
        raise NotImplementedError

    def __getattr__(self, name: str):
        """See if ``name`` is a custom field."""
        if name in self.custom_fields:
            return self._custom_fields[name]
        else:
            raise AttributeError from None

    def __setattr__(self, name, value):
        """Set custom custom_fields if a valid key."""
        if name in self.custom_fields:
            self._custom_fields[name] = value
        else:
            super().__setattr__(name, value)

    def __lt__(self, other):
        """Library items implement the `lt` magic method to allow sorting."""
        raise NotImplementedError


class PathType(sa.types.TypeDecorator):
    """A custom type for paths for database storage.

    Normally, paths are Path type, but we can't store that in the database,
    so we normalize the paths first to strings for database storage. Paths are stored as
    relative paths from ``library_path`` in the config.
    """

    impl = sa.types.String  # sql type
    cache_ok = True  # expected to produce same bind/result behavior and sql generation

    library_path: Path  # will be set on config initialization

    def process_bind_param(self, pathlib_path, dialect):
        """Normalize pathlib paths as strings for the database.

        Args:
            pathlib_path: Inbound path to the db.
            dialect: Database in use.

        Returns:
            Relative path from ``library_path`` if possible, otherwise stores the
            absolute path.
        """
        try:
            return str(pathlib_path.relative_to(self.library_path))
        except ValueError:
            return str(pathlib_path.resolve())

    def process_result_value(self, path_str, dialect):
        """Convert the path back to a Path object on the way out."""
        if path_str is None:
            return None

        return Path(self.library_path / path_str)
